from setuptools import setup

setup(name='pydsi',
      version='0.1.2',
      description='Statistical variational inference for dynamical systems.',
      long_description='',
      keywords='variational data assimilation chaos dynamical systems nonlinear',
      url='http://github.com/joetaylor94/pydsi',
      author='Joseph D. Taylor',
      author_email='j.taylor@bath.ac.uk',
      license='MIT',
      packages=['pydsi'],
      include_package_data=True,
      zip_safe=False)
